from datetime import datetime, timezone, timedelta
import os
import queue
import threading
import time
import json
import signal
import sys
import traceback
import argparse
import csv
import dbf
import pathlib


from concurrent.futures import ThreadPoolExecutor
from pymongo import MongoClient, ASCENDING, DESCENDING

from ..mongo_client_gateway import MongoClientTradeGateway
from ..utils import get_exchange_from_ticker
from .constants import (ENCODING, FILES, GATEWAY_NAME)
from .utils import (decode_atx_status,side_to_target_type, decode_exchange_id, decode_atx_side, decode_ordtype)
from .logger import Logger


try:
    import thread
except ImportError:
    import _thread as thread

LL9 = 1000000000


class atxServer(MongoClientTradeGateway):
    def __init__(self, config_filename, endtime):
        MongoClientTradeGateway.__init__(self, config_filename, endtime, GATEWAY_NAME)

        self.load_tg_setting(config_filename)
        self.order_lock = threading.Lock()
        self.trade_lock = threading.Lock()
        self.pos_lock = threading.Lock()
        self.acc_lock = threading.Lock()
        
        self.insert_orders = {}
        
        self.oid_to_ref = {}
        self.sids = []
        self.oid_to_traded = {}
        self.oid_to_traded_money = {}
        
        self.sid_to_req = {}
        self.oid_to_req = {}
        self.oid_to_mid = {}
        self.oid_to_acc = {}
        self.db_id_to_oid = {}
        self.ref_to_oid = {}
        
        self.order_index = -1 #dbf文件当前index
        self.trade_index = -1
        self.oid_to_local_ids = {}
        
    def monitor(self):
        self.thread_pool.submit(self.monitor_algo_order_insert)
        self.thread_pool.submit(self.monitor_cancel_order_insert)
        self.thread_pool.submit(self.monitor_order_update)
        self.thread_pool.submit(self.monitor_trade_update)
        self.thread_pool.submit(self.date_change)
        #self.thread_pool.submit(self.monitor_dbf_asset)
        self.thread_pool.submit(self.monitor_dbf_pos)
        #self.thread_pool.submit(self.monitor_pos_update)
        #self.thread_pool.submit(self.monitor_acc_update)
        
    def load_tg_setting(self, config_filename):
        try:
            #config_filename = os.path.join(config_path, 'atx_server_config.json')
            f = open(config_filename, encoding='gbk')
            setting = json.load(f)
            #self.insert_order_path = setting['insert_order_path'].replace('/', '\\')
            self.recv_msg_dir = setting['recv_msg_path'].replace('/', '\\')
    
        except Exception as e:
            err = traceback.format_exc()
            exc_type, exc_value, exc_traceback = sys.exc_info()
            traceback.print_exception(exc_type, exc_value, exc_traceback, limit=None, file=sys.stdout)
            print(f"load config failed! (exception){err}")
            exit(0)
    
    
    def init_dbf_tables(self):

        cancel_filename = self.recv_msg_dir + '\\' + FILES.CANCEL + self.date + '.dbf'
        self.cancel_order_table = dbf.Table(filename=cancel_filename,codepage='utf8', on_disk=True)
        self.cancel_order_table.open(mode=dbf.READ_WRITE)
        self.logger.info(f"[init_dbf_tables open] (filename){cancel_filename}")
        #not use
        pos_filename = self.recv_msg_dir + '\\' + FILES.POSITIONS + self.date + '.dbf'
        
        self.position_order_table = dbf.Table(filename=pos_filename,codepage='utf8', on_disk=True)
        self.position_order_table.open(mode=dbf.READ_WRITE)
        self.pos_index = -1
        self.logger.info(f"[init_dbf_tables open] (filename){pos_filename}")
    
    @MongoClientTradeGateway.error_handler
    def str2bool(self, v:str):
        if isinstance(v, bool):
            return v
        if v.lower() in ('yes', 'true', 't', 'y', '1'):
            return True
        elif v.lower() in ('no', 'false', 'f', 'n', '0'):
            return False
        else:
            raise argparse.ArgumentTypeError('Boolean value expected.')
    
    @MongoClientTradeGateway.error_handler
    def req_buy_order_insert(self, obj):
        
        target_account_name = obj['accountName']
        acc = self.target_account_names_to_acc[target_account_name]
        account_id = self.account_id[acc]
        
        oid = str(self.gen_order_id())
        mid = str(obj['mid'])
        vol =  int(obj["volume"])
        basket_name = mid
        #order_vol = (str)(obj["volume"])
        price = 0.0
        if not obj.get('price') is None:
            price = float(obj['price'])
        price_type = 0 #限价
        if not obj.get('price_type') is None:
            price_type = int(obj['price_type'])
        financial_buy_s = obj["financingBuy"]
        side = -1
        if financial_buy_s == True:
            side = 5
        else:
            side = 1

        order_vol = f'{vol:d}'
        int_stock_code = int(obj['ticker'])
        stock_code = get_exchange_from_ticker(int_stock_code)
        algo_type = decode_ordtype(obj['executionPlan']["order_type"])
        begin_time  = self.date + str(obj['executionPlan']["start_time"]).replace(':','') + "000"

        end_time = self.date + str(obj['executionPlan']["end_time"]).replace(':','') + "000"
        
        ExternalId = oid
        LimAction = 1  #涨跌停后交易
        if not obj.get('LimAction') is None :
            if not LimAction:
                LimAction = 0

        AftAction = 0  #时间结束后不交易
        participate = 10
        if not obj.get('participate') is None:
            participate = int(obj['participate'])
        trade_param = f"basket_id={oid}"
        if algo_type == 105:
            trade_param += f':maxpercentageF={participate}'
        elif algo_type == 201:
            trade_param += f':PriceTypeI={price_type}:priceF={price}'

        insert_row = (ExternalId, account_id, stock_code, side, order_vol, algo_type, begin_time, end_time, LimAction, AftAction, trade_param)
        #insert_row = (ExternalId, client_name, stock_code, side, order_vol, algo_type_using, begin_time, end_time, LimAction, AftAction, trade_param)
        self.insert_order_table.append(insert_row)
            
        self.logger.info(f"[req_buy_order] insert_success (oid){oid} (mid){mid} (param){insert_row}")
        db_id = obj['_id']
        order_dict = {
            'oid' : oid,
            'mid' : mid,
            "db_id" : db_id,
            "ticker" : obj['ticker'],
            "side" : side,
            "volume" : order_vol,
            "accountName" : target_account_name,
            "order_type" : str(obj['executionPlan']["order_type"]),
            "start_time" : str(obj['executionPlan']["start_time"]),
            "end_time" : str(obj['executionPlan']["end_time"])
        }
        
        target_collection =  self.order_info_db[acc]['target']
        delete_query = {
                    '_id' : db_id
                }
        delete_res = target_collection.delete_one(delete_query)
        self.logger.info(f"[on_req_buy_order] delete (target){obj}")
        
        atx_order_collection = self.order_info_db[acc]['atx_order']
        local_id_list = []
        db_msg = {
                    "oid" : oid,
                    "mid" : mid,
                    "traded_vol" : 0,
                    "traded_amt" : 0,
                    "local_ids" : local_id_list,
                    "order_msg": order_dict
                }
        query = {"oid": oid}
        res = atx_order_collection.replace_one(query, db_msg, True)
        self.logger.info(f"[on_req_order_insert] insert_atx_order_info (res){res} (msg){db_msg}")

        self.sids.append(mid)

        self.sid_to_req[mid] = order_dict
        self.oid_to_req[oid] = order_dict
        self.oid_to_mid[oid] = mid
        self.db_id_to_oid[db_id] = oid
        self.oid_to_acc[oid] = acc
        self.oid_to_traded[oid] = 0
        self.oid_to_traded_money[oid] = 0
        self.oid_to_local_ids[oid] = local_id_list

    @MongoClientTradeGateway.error_handler
    def req_sell_order_insert(self, obj):
        target_account_name = obj['accountName']
        acc = self.target_account_names_to_acc[target_account_name]
        account_id = self.account_id[acc]

        oid = str(self.gen_order_id())
        mid = str(obj['mid'])
        vol =  int(obj["volume"])
        #order_vol = (str)(obj["volume"])
        #financial_buy_s = self.str2bool(str(obj["financingBuy"]))
        side = 2
        price = 0.0
        if not obj.get('price') is None:
            price = float(obj['price'])
        price_type = 0 #限价
        if not obj.get('price_type') is None:
            price_type = int(obj['price_type'])
        order_vol = f'{vol:d}'
        int_stock_code = int(obj['ticker'])
        stock_code = get_exchange_from_ticker(int_stock_code)

        algo_type = decode_ordtype(obj['executionPlan']["order_type"]) 
        algo_type_using = 204  #ft vwap plus
        begin_time  = self.date + str(obj['executionPlan']["start_time"]).replace(':','') + "000"

        end_time = self.date + str(obj['executionPlan']["end_time"]).replace(':','') + "000"

        ExternalId = oid
        LimAction = 1  #涨跌停后交易
        if not obj.get('LimAction') is None :
            if not LimAction:
                LimAction = 0
        AftAction = 0  #时间结束后不交易
        participate = 10
        if not obj.get('participate') is None:
            participate = int(obj['participate'])
        trade_param = f"basket_id={oid}"
        if algo_type == 105:
            trade_param += f':maxpercentageF={participate}'
        elif algo_type == 201:
            trade_param += f':PriceTypeI={price_type}:priceF={price}'

        insert_row = (ExternalId, account_id, stock_code, side, order_vol, algo_type, begin_time, end_time, LimAction, AftAction, trade_param)
        self.insert_order_table.append(insert_row)

        self.logger.info(f"[req_sell_order] insert_success (oid){oid} (mid){mid} (param){insert_row}")
        db_id = obj['_id']
        order_dict = {
            'oid' : oid,
            'mid' : mid,
            "db_id" : db_id,
            "ticker" : obj['ticker'],
            "trade_acc" : acc,
            "side" : side,
            "volume" : order_vol,
            "accountName" : target_account_name,
            "order_type" : str(obj['executionPlan']["order_type"]),
            "start_time" : str(obj['executionPlan']["start_time"]),
            "end_time" : str(obj['executionPlan']["end_time"])
        }

        target_collection =  self.order_info_db[acc]['sell_target']
        delete_query = {
                    "_id" : db_id
                }
        delete_res = target_collection.delete_one(delete_query)
        self.logger.info(f"[on_req_sell_order] delete (target){obj}")

        atx_order_collection = self.order_info_db[acc]['atx_order']
        local_id_list = []
        db_msg = {
                    "oid" : oid,
                    "mid" : mid,
                    "traded_vol" : 0,
                    "traded_amt" : 0,
                    "local_ids" : local_id_list,
                    "order_msg": order_dict
                }
        query = {"oid": oid}
        res = atx_order_collection.replace_one(query, db_msg, True)
        self.logger.info(f"[on_req_order_insert] insert_sell_order_info (res){res} (msg){db_msg}")

        self.sids.append(mid)
        self.sid_to_req[mid] = order_dict
        self.oid_to_req[oid] = order_dict
        self.oid_to_mid[oid] = mid
        self.db_id_to_oid[db_id] = oid
        self.oid_to_acc[oid] = acc
        self.oid_to_traded[oid] = 0
        self.oid_to_traded_money[oid] = 0
        self.oid_to_local_ids[oid] = local_id_list

    #order_insert 和 rtn_order 都在此获取
    #0 已报 #2 filled 全成 #4 canceled #8 rejected
    @MongoClientTradeGateway.error_handler
    def monitor_order_update(self):
        print ("[zghs_atx_monitor_order_update]")
        while not self.is_stopped:
            with self.order_lock:
                order_filename = self.recv_msg_dir + '\\' + FILES.ORDERS + self.date + '.dbf'
                rtn_order_table = dbf.Table(filename=order_filename,codepage='utf8', on_disk=True)
                rtn_order_table.open(mode=dbf.READ_WRITE)
                self.order_index = -1
                if self.order_index == len(rtn_order_table)-1 or len(rtn_order_table) == 0:
                    self.logger.info(f"[monitor_order_update] not_change (len){len(rtn_order_table)}")
                    order_scan_interval = self.order_scan_interval / 2
                    time.sleep(order_scan_interval)
                    continue
                else:
                    while rtn_order_table[self.order_index+1] is not rtn_order_table.last_record:
                        self.order_index += 1
                        record = rtn_order_table[self.order_index]
                        if record.OrdStatus == 8:
                            self.logger.warning(f"[monitor_order_update] order_rejected! (msg){str(record.Text).strip(' ')}")
                            #self.logger.warn(f"[monitor_order_update] order_rejected!")
                        elif record.OrdStatus in [0,2,4]:
                            self.update_order(record)
                        else:
                            self.logger.info("[monitor_order_update] order_not_finished_or_not_init")

                    if rtn_order_table.last_record.OrdStatus in [0,2,4]:
                        self.order_index += 1
                        self.update_order(rtn_order_table.last_record)
                rtn_order_table.close()
            time.sleep(self.order_scan_interval)
    
    @MongoClientTradeGateway.error_handler
    def update_order(self,record):
        oid = str(record.ExternalId).strip(' ')
        acc = ""
        mid = ""
        order_dict = {}
        mudan_id = str(record.QuoteId).strip(' ')
        if not oid in self.oid_to_acc: #shut_down or other problem
            trade_acc = str(record.ClientName).strip(' ')
            if trade_acc not in self.account_id_to_acc:
                self.logger.error(f"[update_order] can't_parse_trade_acc {trade_acc}")
                return
            acc = self.account_id_to_acc[trade_acc]
            order_info_collection = self.order_info_db[acc]['atx_order']
            query = {'oid' : oid}
            order_info_target = order_info_collection.find_one(query)
            if not order_info_target is None:
                order_dict = order_info_target['order_msg']
                mid = order_info_target['mid']
                self.oid_to_traded_money[oid] = 0
                self.oid_to_traded[oid] = 0
                self.oid_to_mid[oid] = mid
                self.oid_to_req[oid] = mid
                self.oid_to_ref[oid] = mudan_id
            else:
                self.logger.error(f"[update_order] can't_find_oid (oid){oid}")
                return
        else:
            acc = self.oid_to_acc[oid]
            order_dict = self.oid_to_req[oid]
            mid = self.oid_to_mid[oid]
        ticker = record.Symbol.split('.')[0]
        exchange = decode_exchange_id(str(record.Symbol.split('.')[1]).strip(' '))
        volume = record.OrderQty
        price = record.AvgPx

        tg_name = order_dict['accountName'].split('@')[0]
        accountName = order_dict['accountName'].split('@')[1]
        order_type = order_dict['order_type']
        start_time = order_dict['start_time']
        end_time = order_dict['end_time']

        target_type = side_to_target_type(order_dict['side'])
        filled_vol = record.CumQty
        status = decode_atx_status(record.OrdStatus)
        
        replace_collection = self.tradelog_db[acc]['order']
        query = {'mid':mid, '_id': int(oid)}
        replace_target = replace_collection.find_one(query)
        if not replace_target is None:
        
            #str_update_time = str(int(int(record.UpdateTime)/1000))
            #update_time = datetime.strptime(str_update_time, "%Y%m%d%H%M%S")
            #utc_update_time = update_time - timedelta(hours=8)
            utc_update_time = replace_target['dbTime']

            self.oid_to_ref[oid] = mudan_id
            order_msg = {
                    "_id": int(oid),
                    "tg_name": tg_name,  # 对应 account_info._id
                    # 对应 tlclient.trader.constant 包 ExchangeID
                    "exchange": exchange,
                    "target_type": target_type,    # 'buy' | 'sell' | 'limit_sell'
                    "volume": volume,  # 订单的volume
                    "price": price,  # 实际成交均价
                    "order_type": order_type,  # algotype:200,目前没有其他的
                    "ticker": ticker,
                    "mid": mid,  # target中对应的 母单sid
                    "accountName": accountName,  # 对应 account_info.account_name
                    "algo_args": {  # 具体的算法参数
                        "order_type": order_type,  # 此柜台不使用下单时要求的算法
                        "start_time": start_time,
                        "end_time": end_time
                    },
                    "status": status,  # 'active' | 'filled' | 'canceled'
                    "filled_vol": filled_vol,  # 实际成交的 volume
                    "dbTime": utc_update_time,
            }
            
            res = replace_collection.replace_one(query, order_msg, True)
            self.logger.info(f"[rtn_order] (res){res} (order_msg){order_msg}")
        else:
            str_update_time = str(int(record.TRANSTIME))
            update_time = datetime.strptime(str_update_time, "%Y%m%d%H%M%S%f")
            utc_update_time = update_time - timedelta(hours=8)
            #utc_update_time = replace_target['dbTime']

            self.oid_to_ref[oid] = mudan_id
            order_msg = {
                    "_id": int(oid),
                    "tg_name": tg_name,  # 对应 account_info._id
                    # 对应 tlclient.trader.constant 包 ExchangeID
                    "exchange": exchange,
                    "target_type": target_type,    # 'buy' | 'sell' | 'limit_sell'
                    "volume": volume,  # 订单的volume
                    "price": price,  # 实际成交均价
                    "order_type": order_type,  # algotype:200,目前没有其他的
                    "ticker": ticker,
                    "mid": mid,  # target中对应的 母单sid
                    "accountName": accountName,  # 对应 account_info.account_name
                    "algo_args": {  # 具体的算法参数
                        "order_type": order_type,  # 此柜台不使用下单时要求的算法
                        "start_time": start_time,
                        "end_time": end_time
                    },
                    "status": status,  # 'active' | 'filled' | 'canceled'
                    "filled_vol": filled_vol,  # 实际成交的 volume
                    "dbTime": utc_update_time,
            }
            
            res = replace_collection.insert_one(order_msg)
            self.logger.info(f"[rtn_order] (res){res} (order_msg){order_msg}")
       
    
    @MongoClientTradeGateway.error_handler
    def monitor_algo_order_insert(self):
        while not self.is_stopped:
            print ("[monitor_algo_order_insert]")
            insert_filename = self.recv_msg_dir + '\\' + FILES.TASKS + self.date + '.dbf'
            self.insert_order_table = dbf.Table(filename=insert_filename,codepage='utf8', on_disk=True)
            with self.insert_order_table.open(mode=dbf.READ_WRITE):
                self.logger.info(f"[init_dbf_tables open] (filename){insert_filename}")
                for acc in self.accounts_run:
                    target_account_name = self.target_account_names[acc]
                    buy_query = {"accountName": target_account_name}
                    buy_targets = self.order_info_db[acc]["target"].find(buy_query)
                    if buy_targets.count() == 0:
                        self.logger.warning(f"[monitor_buy_order] no_buy_target (acc){acc}")
                        continue
                    for target in buy_targets:
                        if target['_id'] not in self.order_db_ids:
                            self.order_db_ids.append(target['_id'])
                            self.req_buy_order_insert(target)
                        else:
                            self.logger.warning(f"[monitor_algo_order_insert] _id_existed (_id){target['_id']}")
                for acc in self.accounts_run:
                    target_account_name = self.target_account_names[acc]
                    sell_query = {"accountName": target_account_name}
                    sell_targets = self.order_info_db[acc]["sell_target"].find(sell_query)
                    if sell_targets.count() == 0:
                        self.logger.warning(
                            f"[monitor_sell_order] no_sell_target (acc){acc}")
                        continue
                    for sell_target in sell_targets:
                        if sell_target['_id'] not in self.sell_order_db_ids:
                            self.sell_order_db_ids.append(sell_target['_id'])
                            self.req_sell_order_insert(sell_target)
                        else:
                            self.logger.warning(f"[monitor_algo_order_insert] _id_existed (_id){sell_target['_id']}")
            time.sleep(self.scan_interval)
    
    @MongoClientTradeGateway.error_handler
    def monitor_cancel_order_insert(self):
        while not self.is_stopped:
                for acc in self.accounts_run:
                    print ("[monitor_cancel_order_insert]")
                    target_account_name = self.target_account_names[acc]
                    query = {"accountName": target_account_name}
                    with self.cancel_orderlock:
                        cancel_targets = self.order_info_db[acc]["cancel_target"].find(
                            query)
                        if cancel_targets.count() == 0:
                            self.logger.warning(
                                "[monitor_cancel_order] no_cancel_info")
                            continue
                        for cancel_target in cancel_targets:
                            if cancel_target['_id'] not in self.cancel_order_ids:
                                self.cancel_order_ids.append(cancel_target['_id'])
                                self.req_cancel_order(cancel_target)
                time.sleep(self.scan_interval)

    @MongoClientTradeGateway.error_handler
    def req_cancel_order(self, cancel_target):
        oid = str(cancel_target['oid'])
        mudan_id = ""
        if oid not in self.oid_to_ref:
           self.logger.error("[req_cancel_order] can't_find_mudanid (oid){oid}")
           return
        mudan_id = self.oid_to_ref[oid]
        cxltype = 1
        row = [mudan_id, cxltype]
        self.cancel_order_table.append(row)
        
        origin_req = self.oid_to_req[oid]
        target_account_name = origin_req['accountName']
        acc = self.target_account_names_to_acc[target_account_name]
        target_collection =  self.order_info_db[acc]['cancel_target']
        delete_query = {
            'accountName' : target_account_name,
            'oid' : oid
            }
        delete_res = target_collection.delete_one(delete_query)
        self.logger.info(f"[on_req_cancel_order] delete (target){cancel_target}")
    
    #得到的数据 类似于子单的rtn_order
    @MongoClientTradeGateway.error_handler
    def monitor_trade_update(self):
        while not self.is_stopped:
            with self.trade_lock:
                trade_filename = self.recv_msg_dir + '\\' + FILES.TRADES + self.date + '.dbf'

                rtn_trade_table = dbf.Table(filename=trade_filename,codepage='utf8', on_disk=True)
                rtn_trade_table.open(mode=dbf.READ_WRITE)
                self.trade_index = -1
                if self.trade_index == len(rtn_trade_table)-1 or len(rtn_trade_table) == 0:
                    self.logger.info(f"[monitor_trade_update] not_change (len){len(rtn_trade_table)}")
                    trade_scan_interval = self.trade_scan_interval / 2
                    time.sleep(trade_scan_interval)
                    continue
                else:
                    while rtn_trade_table[self.trade_index+1] is not rtn_trade_table.last_record:
                        self.trade_index += 1
                        record = rtn_trade_table[self.trade_index]
                        if record.OrdStatus in [2,4]: #only need filled/canceled
                            self.on_trade_msg(record)
                        else:
                            self.logger.info(f"[monitor_trade_update] traded_nofinished (msg){record.Text}")

                    self.trade_index += 1
                    record = rtn_trade_table[self.trade_index]
                    if record.OrdStatus in [2,4]: #only need filled/canceled
                        self.on_trade_msg(record)
                    else:
                        self.logger.info(f"[monitor_trade_update] traded_nofinished (msg){record.Text}")
                    

                rtn_trade_table.close()
            time.sleep(self.trade_scan_interval)

    @MongoClientTradeGateway.error_handler
    def on_trade_msg(self, record):
        if record.CumQty == 0 or record.AvgPx == 0:
            self.logger.info(f"[on_trade_msg] nothing_traded")
            return
        oid = str(record.ExternalId).strip(' ')
        mudan_id = str(record.QuoteId).strip(' ')
        acc = ""
        mid = ""
        order_dict = {}
        if not oid in self.oid_to_acc: #shut_down or other problem
            trade_acc = str(record.ClientName).strip(' ')
            acc = self.account_id_to_acc[trade_acc]
            order_info_collection = self.order_info_db[acc]['atx_order']
            query = {'oid' : oid}
            order_info_target = order_info_collection.find_one(query)

            if not order_info_target is None:
                order_dict = order_info_target['order_msg']
                mid = order_info_target['mid']
                self.oid_to_local_ids[oid] = order_info_target['local_ids']
                self.oid_to_traded[oid] = 0
                self.oid_to_traded_money[oid] = 0
                self.oid_to_mid[oid] = mid
                self.oid_to_req[oid] = order_dict
                self.oid_to_ref[oid] = mudan_id
            else:
                self.logger.error(f"[on_trade_msg] can't_find_oid (oid){oid}")
                return
        else:
            acc = self.oid_to_acc[oid]
            order_dict = self.oid_to_req[oid]
            mid = self.oid_to_mid[oid]
        _id = str(record.ClOrdId).strip(' ')
        if _id in self.oid_to_local_ids[oid]:
            #self.logger.warning(f"[on_trade_msg] duplicate_trade_msg (record){record}")
            return
        else:
            self.oid_to_local_ids[oid].append(_id)
        trade_ref = str(record.OrderId).strip(' ') #外部委托编号，出现在这个位置主要是凑数
        ticker = str(record.Symbol).split('.')[0]
        exchange = decode_exchange_id(str(record.Symbol).split('.')[1].strip(' '))
        traded_vol = record.CumQty
        traded_price = record.AvgPx
        trade_amt = float(traded_price * traded_vol)
        order_type = record.OrdType
        target_type = decode_atx_side(record.Side)
        if target_type == 1 or target_type == 3:
            self.oid_to_traded_money[oid] += trade_amt
            self.oid_to_traded[oid] += traded_vol
        elif target_type == 2:
            self.oid_to_traded_money[oid] -= trade_amt
            self.oid_to_traded[oid] -= traded_vol
        entrust_vol = record.OrderQty
        #entrust_price = record.Price
        entrust_price = float(str(record.Price).strip(' '))
        dbTime = datetime.now(timezone.utc)
        
        commission = 0
        str_trade_time = str(record.UpdateTime)
        trade_time = datetime.strptime(str_trade_time, "%Y%m%d%H%M%S%f")
        utc_trade_time = trade_time - timedelta(hours=8)
        target_account_name = order_dict['accountName']
        acc = self.target_account_names_to_acc[target_account_name]
        log_account_name = order_dict['accountName'].split('@')[1]
        tg_name =  order_dict['accountName'].split('@')[0]
        side = decode_atx_side(record.Side)
        trade_collection = self.tradelog_db[acc]['tg_trade']
        replace_trade_query = { "trade_ref": _id, "oid" : int(oid), "mid": mid, "accountName": log_account_name}
        trade_target = trade_collection.find_one(replace_trade_query)
        if not trade_target is None:
            dbTime = trade_target['dbTime']

        db_msg = {
                    "trade_ref": _id, # 用的是非凸的子单order_id 经测试之前用的trade_ref 经常为空
                    "oid": int(oid),
                    "tg_name": tg_name,
                    "exchange": exchange,
                    "ticker": ticker,
                    "traded_vol": traded_vol,
                    "traded_price": traded_price,
                    "order_type": order_type,
                    "side": side,  # 对应 tlclient.trader.constant 包 Side
                    "entrust_vol": entrust_vol,
                    "entrust_price": entrust_price,
                    "dbTime": dbTime,
                    "mid": mid,  # 对应订单中的 mid
                    "commission": 0,  # 没有
                    "trade_time": utc_trade_time,  # 具体交易时间
                    "accountName":  log_account_name,  # 对应 account_info.account_name
                }
        db_res = trade_collection.replace_one(replace_trade_query, db_msg, True)
        self.logger.info(
            f"[rtn_trade] (db_res){db_res} (db_msg){db_msg} (traded_vol){traded_vol} (traded_price){traded_price}")
        #self.update_position(traded_vol, order_dict, mid, oid, side, trade_amt,exchange)

        atx_order_collection = self.order_info_db[acc]['atx_order']
        atx_order_msg = {
                    "local_ids" : self.oid_to_local_ids[oid],
                }
        update_atx_order_msg = {"$set": atx_order_msg}
        query = {"oid": oid}
        res = atx_order_collection.update_one(query, update_atx_order_msg)
        self.logger.info(f"[on_trade_msg] update_atx_order_info (res){res} (msg){update_atx_order_msg}")

    @MongoClientTradeGateway.error_handler
    def update_position(self, traded_vol, order, mid , oid, side, trade_amt, exchange):
        target_account_name = order['accountName']
        acc = self.target_account_names_to_acc[target_account_name]

        query = {'mid': mid, 'accountName': target_account_name}
        position_collection = self.order_info_db[acc]['EquityPosition']
        position = position_collection.find_one(query)
        if position is None and traded_vol > 0:
            if oid not in self.oid_to_req:
                self.logger.warning(
                    f"[update_position] can't_find_req (mid){mid} (oid){oid}")
                return
            if oid not in self.oid_to_traded:
                self.logger.error(f"[update_position]no trade_vol (mid){mid}")
                return
            td_trade_vol = self.oid_to_traded[oid]

            amt = self.oid_to_traded_money[oid]

            ticker = order['ticker']
            yd_pos = 0
            mkt_value = amt
            cost = 0
            if not td_trade_vol == 0:
                cost = amt / td_trade_vol
            update_time = datetime.now(timezone.utc)

            position_msg = {
                "ticker": ticker,
                "cost": cost,  # 成本价
                "td_pos_long": td_trade_vol,  # 今仓
                "yd_pos_long": yd_pos,  # 昨仓（可卖），挂止盈单后为0
                "td_pos_short": 0,
                "yd_pos_short": 0,
                "actual_td_pos_long": td_trade_vol,  # 实际的今仓
                "actual_yd_pos_long": yd_pos,  # 实际的昨仓
                "enter_date": update_time,
                "holding_days": 0,  # 持仓时间（交易日累计）
                "mkt_value": mkt_value,  # 市值
                "exchange": exchange,# 对应 tlclient.trader.constant 包 ExchangeID
                "mid": mid,  # 我们策略生成的id，需要根据实际下单情况维护各sid的持仓。
                # 格式: {account_info.tg_name}@{account_info.account_name}
                "accountName": target_account_name,
                "update_date": update_time,
            }
            query2 = {'mid': mid,
                        'accountName': target_account_name}
            res = position_collection.replace_one(query2, position_msg, True)
            self.logger.info(
                f"[update_position] (res){res} (msg){position_msg}")
        else:
            if oid not in self.oid_to_req:
                self.logger.warning(
                    f"[update_position] can't_find_req (mid){mid}")
                return
            if oid not in self.oid_to_traded:
                self.logger.error(f"[update_position]no trade_vol (mid){mid}")
                return
            actual_yd_pos = position['actual_yd_pos_long']
            amt = trade_amt
            yd_pos = position['yd_pos_long']
            if side == 1 or side == 3:
                td_trade_vol = position['actual_td_pos_long'] + traded_vol
                mkt_value = position['mkt_value'] + amt
            elif side == 2 :
                td_trade_vol = position['actual_td_pos_long'] - traded_vol
                mkt_value = position['mkt_value'] - amt
                yd_pos = yd_pos - traded_vol

            ticker = order['ticker']

            cost = 0
            total_vol = td_trade_vol + yd_pos
            if not total_vol == 0:
                cost = mkt_value / total_vol
            update_time = datetime.now(timezone.utc)
            accountName = order['accountName']
            enter_time = position['enter_date']
            holding_days = position['holding_days']
            position_msg = {
                "ticker": ticker,
                "cost": cost,  # 成本价
                "td_pos_long": td_trade_vol,  # 今仓
                "yd_pos_long": yd_pos,  # 昨仓（可卖），挂止盈单后为0
                "td_pos_short": 0,
                "yd_pos_short": 0,
                "actual_td_pos_long": td_trade_vol,  # 实际的今仓
                "actual_yd_pos_long": actual_yd_pos,  # 实际的昨仓
                "enter_date": enter_time,
                "holding_days": holding_days,  # 持仓时间（交易日累计）
                "mkt_value": mkt_value,  # 市值
                "exchange": exchange,  # 对应 tlclient.trader.constant 包 ExchangeID
                "mid": mid,  # 我们策略生成的id，需要根据实际下单情况维护各sid的持仓。
                # 格式: {account_info.tg_name}@{account_info.account_name}
                "accountName": accountName,
                "update_date": update_time,
            }
            query = {'mid': mid,
                        'accountName': accountName}
            res = position_collection.replace_one(query, position_msg, True)
            self.logger.info(
                f"[update_position] (res){res} (msg){position_msg}")

    def start(self):
        signal.signal(signal.SIGTERM, self.signal_handler)
        signal.signal(signal.SIGINT, self.signal_handler)
        _msg = f"[login] zghs_atx_server_start (time){datetime.now()}"
        self.send_to_user(logger=self.logger, url_list=self.url_list, msg=_msg)
        self.init_dbf_tables()
        self.monitor()

    def close(self):

        self.position_order_table.close()
        self.cancel_order_table.close()
        return super().close()
    
    def join(self):
        while self.is_stopped == False:
            time.sleep(0.01)
            if self.is_stopped:
                self.logger.info(
                    "[close] main thread is stopped,active_orders message will lose")
                self.close()

if __name__ == "__main__":

    sys.stdout.reconfigure(encoding='utf-8')
    description = "atx_server,get target from mongodb and serve atx"  
    parser = argparse.ArgumentParser(description=description)

    parser.add_argument('-e' , '--end_time', dest='end_time', default='15:30')
    _config_filename = "C:/Users/Administrator/Desktop/zghs_atx_batandconfig/zghs_atx_server_config.json"
    parser.add_argument('-p', '--config_filepath', dest= 'config_filepath', default= _config_filename)

    args = parser.parse_args()
    print (f"(args){args}")

    td = atxServer(args.config_filepath, args.end_time)
    td.start()

    td.join()


from datetime import datetime, timezone, timedelta
import os
import queue
import threading
import time
import json
import signal
import sys
import traceback
import argparse
import csv
import dbf
import pathlib
from openpyxl import Workbook
import shutil

from concurrent.futures import ThreadPoolExecutor
from pymongo import MongoClient, ASCENDING, DESCENDING

from mongo_client_gateway import MongoClientTradeGateway

from constants import (ENCODING, FILES, Exchange, OrderStatus, OrderType, SecurityType, Side)
from utils import (decode_ims_status,side_to_target_type, decode_exchange_id, encode_ims_side, decode_ordtype, decode_ims_side, decode_ims_contract_type, decode_ims_market, check_charset)

from logger import Logger


try:
    import thread
except ImportError:
    import _thread as thread

LL9 = 1000000000


class ImsServer(MongoClientTradeGateway):
    def __init__(self, config_filename, endtime):
        MongoClientTradeGateway.__init__(self, config_filename, endtime)

        self.load_tg_setting(config_filename)
        self.order_lock = threading.Lock()
        self.trade_lock = threading.Lock()
        self.pos_lock = threading.Lock()
        self.acc_lock = threading.Lock()
        

        self.order_row_set = set()
        self.trade_row_set = set()
        self.pos_row_set = set()
        self.acc_row_set = set()
        self.cancel_row_set = set()

        self.insert_orders = {}
        self.oid_to_ref = {}
        self.sids = []
        self.oid_to_traded = {}
        self.oid_to_traded_money = {}
        
        self.sid_to_req = {}
        self.oid_to_req = {}
        self.oid_to_mid = {}
        self.oid_to_acc = {}
        self.db_id_to_oid = {}
        self.ref_to_oid = {}
        
        self.order_index = -1 #dbf文件当前index
        self.trade_index = -1
        self.oid_to_local_ids = {}
        
    def monitor(self):
        self.thread_pool.submit(self.monitor_algo_order_insert)
        self.thread_pool.submit(self.monitor_cancel_order_insert)
        #self.thread_pool.submit(self.monitor_order_update)

        self.thread_pool.submit(self.monitor_trade_update)

        self.thread_pool.submit(self.date_change)
        #self.thread_pool.submit(self.monitor_pos_update)
        #self.thread_pool.submit(self.monitor_acc_update)
        
    def load_tg_setting(self, config_filename):
        try:
            #config_filename = os.path.join(config_path, 'atx_server_config.json')
            f = open(config_filename, encoding='gbk')
            setting = json.load(f)
            #self.insert_order_path = setting['insert_order_path'].replace('/', '\\')
            self.recv_msg_dir = setting['recv_msg_path'].replace('/', '\\')
            self.insert_order_msg_dir = setting['insert_order_msg_path'].replace('/', '\\')


        except Exception as e:
            err = traceback.format_exc()
            exc_type, exc_value, exc_traceback = sys.exc_info()
            traceback.print_exception(exc_type, exc_value, exc_traceback, limit=None, file=sys.stdout)
            print(f"load config failed! (exception){err}")
            exit(0)

    @MongoClientTradeGateway.error_handler
    def str2bool(self, v:str):
        if isinstance(v, bool):
            return v
        if v.lower() in ('yes', 'true', 't', 'y', '1'):
            return True
        elif v.lower() in ('no', 'false', 'f', 'n', '0'):
            return False
        else:
            raise argparse.ArgumentTypeError('Boolean value expected.')
    
    @MongoClientTradeGateway.error_handler
    def req_buy_order_insert(self, obj):
        workbook = Workbook()

        sheet = workbook.active

        headers = [
            '合约代码',
            '市场',
            '交易类型',
            '投资组合',
            '合约类型',
            '是否市价',
            '指令价格',
            '指令数量',
            '算法策略类型',
            '算法开始时间',
            '算法结束时间',
            '算法参数',
            '备注'
        ]

        for col_num, header in enumerate(headers, 1):
            sheet.cell(row=1, column=col_num).value = header

        target_account_name = obj['accountName']
        acc = self.target_account_names_to_acc[target_account_name]
        account_id = self.account_id[acc]

        oid = str(self.gen_order_id())
        mid = str(obj['mid'])
        vol =  int(obj["volume"])
        side = int(obj["side"])
        # Write some sample data to the rows below

        target_type = decode_ims_side(side)
        basket_name = mid
        order_vol = f'{vol:d}'
        int_stock_code = int(obj['ticker'])
        stock_code = f'{int_stock_code:06d}'
        market = decode_ims_market(stock_code)
        algo_type = decode_ordtype(obj['executionPlan']["order_type"])

        begin_time = str(obj['executionPlan']["start_time"]).replace(":", "")
        end_time = str(obj['executionPlan']["end_time"]).replace(":", "")

        trade_param = ""
        ExternalId = oid
        #contract_code = obj['contract_code']

        combo_id = obj['combo_id']
        int_contract_type = self.contract_type[acc]
        contract_type = decode_ims_contract_type(int_contract_type)
        is_market = obj['is_market']
        price = int(obj['price'])
        str_limit = ""
        if is_market == 0:
            str_limit = "LIMIT"
        elif is_market == 1:
            str_limit = "ANY"
        
        data = [
            [stock_code, market, target_type, combo_id, contract_type, str_limit, price, vol, algo_type, begin_time, end_time, trade_param, ExternalId],
        ]

        for row_num, row_data in enumerate(data, 2):
            for col_num, cell_value in enumerate(row_data, 1):
                sheet.cell(row=row_num, column=col_num).value = cell_value

        temp_filename = self.insert_order_msg_dir + '_temp'+ '\\' + 'buy' + oid + '.xlsx'
        # Save the workbook
        workbook.save(temp_filename)

        workbook.close()

        # Move the temporary file to the final filename
        final_filename = self.insert_order_msg_dir + '\\' + 'buy' + oid + '.xlsx'
        shutil.move(temp_filename, final_filename)

        self.logger.info(f"[req_buy_order] insert_success (oid){oid} (mid){mid} (param){data}")
        db_id = obj['_id']
        order_dict = {
            'oid' : oid,
            'mid' : mid,
            "db_id" : db_id,
            "ticker" : obj['ticker'],
            "side" : side,
            "volume" : order_vol,
            "accountName" : target_account_name,
            "order_type" : str(obj['executionPlan']["order_type"]),
            "start_time" : str(obj['executionPlan']["start_time"]),
            "end_time" : str(obj['executionPlan']["end_time"]),
            "contract_type" : int_contract_type,
            "price" : obj['price'],
            "is_market": obj['is_market'],
            "combo_id" : combo_id
        }
        
        target_collection =  self.order_info_db[acc]['target']
        delete_query = {
                    '_id' : db_id
                }
        delete_res = target_collection.delete_one(delete_query)
        self.logger.info(f"[on_req_buy_order] delete (target){obj}")
        
        atx_order_collection = self.order_info_db[acc]['atx_order']
        local_id_list = []
        db_msg = {
                    "oid" : oid,
                    "mid" : mid,
                    "traded_vol" : 0,
                    "traded_amt" : 0,
                    "local_ids" : local_id_list,
                    "order_msg": order_dict
                }
        query = {"oid": oid}
        res = atx_order_collection.replace_one(query, db_msg, True)
        self.logger.info(f"[on_req_order_insert] insert_atx_order_info (res){res} (msg){db_msg}")

        self.sids.append(mid)

        self.sid_to_req[mid] = order_dict
        self.oid_to_req[oid] = order_dict
        self.oid_to_mid[oid] = mid
        self.db_id_to_oid[db_id] = oid
        self.oid_to_acc[oid] = acc
        self.oid_to_traded[oid] = 0
        self.oid_to_traded_money[oid] = 0
        self.oid_to_local_ids[oid] = local_id_list

    @MongoClientTradeGateway.error_handler
    def req_sell_order_insert(self, obj):
        workbook = Workbook()

        sheet = workbook.active

        headers = [
            '合约代码',
            '市场',
            '交易类型',
            '投资组合',
            '合约类型',
            '是否市价',
            '指令价格',
            '指令数量',
            '算法策略类型',
            '算法开始时间',
            '算法结束时间',
            '算法参数',
            '备注'
        ]

        for col_num, header in enumerate(headers, 1):
            sheet.cell(row=1, column=col_num).value = header

        target_account_name = obj['accountName']
        acc = self.target_account_names_to_acc[target_account_name]
        account_id = self.account_id[acc]

        oid = str(self.gen_order_id())
        mid = str(obj['mid'])
        vol =  int(obj["volume"])
        side = int(obj["side"])
        # Write some sample data to the rows below

        target_type = decode_ims_side(side)
        basket_name = mid
        order_vol = f'{vol:d}'
        int_stock_code = int(obj['ticker'])
        stock_code = f'{int_stock_code:06d}'
        market = decode_ims_market(stock_code)
        algo_type = decode_ordtype(obj['executionPlan']["order_type"])

        begin_time = str(obj['executionPlan']["start_time"]).replace(":", "")
        end_time = str(obj['executionPlan']["end_time"]).replace(":", "")

        trade_param = ""
        ExternalId = oid
        #contract_code = obj['contract_code']

        combo_id = obj['combo_id']
        int_contract_type = self.contract_type[acc]
        contract_type = decode_ims_contract_type(int_contract_type)
        is_market = obj['is_market']
        price = int(obj['price'])
        str_limit = ""
        if is_market == 0:
            str_limit = "LIMIT"
        elif is_market == 1:
            str_limit = "ANY"
        data = [
            [stock_code, market, target_type, combo_id, contract_type, str_limit, price, vol, algo_type, begin_time, end_time, trade_param, ExternalId],
        ]

        for row_num, row_data in enumerate(data, 2):
            for col_num, cell_value in enumerate(row_data, 1):
                sheet.cell(row=row_num, column=col_num).value = cell_value

        temp_filename = self.insert_order_msg_dir + '_temp' + '\\' + 'sell' + oid + '.xlsx'
        # Save the workbook
        workbook.save(temp_filename)
        
        workbook.close()

        # Move the temporary file to the final filename
        final_filename = self.insert_order_msg_dir + '\\' + 'sell' + oid + '.xlsx'
        shutil.move(temp_filename, final_filename)

        self.logger.info(f"[req_sell_order] insert_success (oid){oid} (mid){mid} (param){data}")
        db_id = obj['_id']
        order_dict = {
            'oid' : oid,
            'mid' : mid,
            "db_id" : db_id,
            "ticker" : obj['ticker'],
            "side" : side,
            "volume" : order_vol,
            "accountName" : target_account_name,
            "order_type" : str(obj['executionPlan']["order_type"]),
            "start_time" : str(obj['executionPlan']["start_time"]),
            "end_time" : str(obj['executionPlan']["end_time"]),
            "contract_type" : int_contract_type,
            "price" : obj['price'],
            "is_market": obj['is_market'],
            "combo_id" : combo_id
        }

        target_collection = self.order_info_db[acc]['sell_target']
        delete_query = {
                    "_id" : db_id
                }
        delete_res = target_collection.delete_one(delete_query)
        self.logger.info(f"[on_req_sell_order] delete (target){obj}")

        atx_order_collection = self.order_info_db[acc]['atx_order']
        local_id_list = []
        db_msg = {
                    "oid" : oid,
                    "mid" : mid,
                    "traded_vol" : 0,
                    "traded_amt" : 0,
                    "local_ids" : local_id_list,
                    "order_msg": order_dict
                }
        query = {"oid": oid}
        res = atx_order_collection.replace_one(query, db_msg, True)
        self.logger.info(f"[on_req_order_insert] insert_sell_order_info (res){res} (msg){db_msg}")

        self.sids.append(mid)
        self.sid_to_req[mid] = order_dict
        self.oid_to_req[oid] = order_dict
        self.oid_to_mid[oid] = mid
        self.db_id_to_oid[db_id] = oid
        self.oid_to_acc[oid] = acc
        self.oid_to_traded[oid] = 0
        self.oid_to_traded_money[oid] = 0
        self.oid_to_local_ids[oid] = local_id_list

    #order_insert 和 rtn_order 都在此获取
    #0 已报 #2 filled 全成 #4 canceled #8 rejected
    @MongoClientTradeGateway.error_handler
    def monitor_order_update(self):
        print ("[monitor_order_update]")
        while not self.is_stopped:
            with self.order_lock:
                order_filename = self.recv_msg_dir  + '\\' + FILES.ORDERS + '.txt'
                                # Specify the path to the input text file

                # Create an empty list to store the dictionaries for each row
                data_list = []

                # Open the file in read mode
                with open(order_filename, 'r') as file:
                    # Read each line in the file
                    for line in file:
                        # Skip lines starting with '#'
                        if line.startswith('#'):
                            continue
                        
                        # Remove leading/trailing whitespaces and split the data using '|'
                        values = line.strip().split('|')
                        
                        # Create a dictionary for the row
                        row_dict = {
                            '产品': values[0],
                            '资产单元': values[1],
                            '投资组合': values[2],
                            '资金帐号': values[3],
                            '市场': values[4],
                            '合约代码': values[5],
                            '序列号': values[6],
                            '委托时间': values[7],
                            '合同号': values[8],
                            '委托数量': values[9],
                            '委托价格': values[10],
                            '买卖方向': values[11],
                            '成交数量': values[12],
                            '成交价格': values[13],
                            '成交编号': values[14],
                            '剩余数量': values[15],
                            '成交状态': values[16],
                            '投资顾问': values[17],
                            '投机套保标志': values[18],
                            '开仓平仓标志': values[19],
                            '篮子ID': values[20],
                            '备注': values[21],
                            '成交时间': values[22],
                            '币种': values[23],
                            '算法ID': values[24],
                            '指令ID': values[25],
                            '审核员': values[26],
                            '交易员': values[27],
                            '错误代码': values[28],
                            '错误消息': values[29],
                            '业务类型': values[30],
                            '子业务类型': values[31],
                            'BookCode': values[32]
                        }
                        
                        # Add the row dictionary to the list
                        data_list.append(row_dict)

                # Print the data list (optional)
                for row_dict in data_list:

                    if row_dict['成交状态'] in ['F_FILLED', 'CANCELED']:
                        self.update_order(row_dict)
                    else:
                        self.logger.info("[monitor_order_update] order_not_finished_or_not_init")


            time.sleep(20)
    
    @MongoClientTradeGateway.error_handler
    def update_order(self,record):
        oid = str(record['备注'])
        acc = ""
        mid = ""
        order_dict = {}
        mudan_id = str(record['投资组合Id'])
        if not oid in self.oid_to_acc: #shut_down or other problem
            trade_acc = str(record['资产单元Id'])

            if trade_acc not in self.account_id_to_acc:
                self.logger.error(f"[update_order] can't_parse_trade_acc {trade_acc}")
                return
            acc = self.account_id_to_acc[trade_acc]
            order_info_collection = self.order_info_db[acc]['atx_order']
            query = {'oid' : oid}
            order_info_target = order_info_collection.find_one(query)
            if not order_info_target is None:
                order_dict = order_info_target['order_msg']
                mid = order_info_target['mid']
                self.oid_to_traded_money[oid] = 0
                self.oid_to_traded[oid] = 0
                self.oid_to_mid[oid] = mid
                self.oid_to_req[oid] = mid
                self.oid_to_ref[oid] = mudan_id
            else:
                self.logger.error(f"[update_order] can't_find_oid (oid){oid}")
                return
        else:
            acc = self.oid_to_acc[oid]
            order_dict = self.oid_to_req[oid]
            mid = self.oid_to_mid[oid]
        ticker = str(record['证券代码'])
        exchange = decode_exchange_id(ticker)
        volume = record['指令数量']
        traded_vol = record['成交数量']
        price = int(record['成交金额']) / int(traded_vol)


        tg_name = order_dict['accountName'].split('@')[0]
        accountName = order_dict['accountName'].split('@')[1]
        order_type = order_dict['order_type']
        start_time = order_dict['start_time']
        end_time = order_dict['end_time']

        target_type = side_to_target_type(order_dict['side'])
        filled_vol = record['成交数量']
        status = decode_ims_status(record['成交状态'])
        
        replace_collection = self.tradelog_db[acc]['order']
        query = {'mid':mid, '_id': int(oid)}
        replace_target = replace_collection.find_one(query)
        if not replace_target is None:

            utc_update_time = replace_target['dbTime']

            self.oid_to_ref[oid] = mudan_id
            order_msg = {
                    "_id": int(oid),
                    "tg_name": tg_name,  # 对应 account_info._id
                    # 对应 tlclient.trader.constant 包 ExchangeID
                    "exchange": exchange,
                    "target_type": target_type,    # 'buy' | 'sell' | 'limit_sell'
                    "volume": volume,  # 订单的volume
                    "price": price,  # 实际成交均价
                    "order_type": order_type,  # algotype:200,目前没有其他的
                    "ticker": ticker,
                    "mid": mid,  # target中对应的 母单sid
                    "accountName": accountName,  # 对应 account_info.account_name
                    "algo_args": {  # 具体的算法参数
                        "order_type": order_type,  # 此柜台不使用下单时要求的算法
                        "start_time": start_time,
                        "end_time": end_time
                    },
                    "status": status,  # 'active' | 'filled' | 'canceled'
                    "filled_vol": filled_vol,  # 实际成交的 volume
                    "dbTime": utc_update_time,
            }
            
            res = replace_collection.replace_one(query, order_msg, True)
            self.logger.info(f"[rtn_order] (res){res} (order_msg){order_msg}")
        else:
            str_update_time = str(record['下单时间'])
            update_time = datetime.strptime(str_update_time, "%Y-%m-%d %H:%M:%S")

            utc_update_time = update_time - timedelta(hours=8)
            #utc_update_time = replace_target['dbTime']

            self.oid_to_ref[oid] = mudan_id
            order_msg = {
                    "_id": int(oid),
                    "tg_name": tg_name,  # 对应 account_info._id
                    # 对应 tlclient.trader.constant 包 ExchangeID
                    "exchange": exchange,
                    "target_type": target_type,    # 'buy' | 'sell' | 'limit_sell'
                    "volume": volume,  # 订单的volume
                    "price": price,  # 实际成交均价
                    "order_type": order_type,  # algotype:200,目前没有其他的
                    "ticker": ticker,
                    "mid": mid,  # target中对应的 母单sid
                    "accountName": accountName,  # 对应 account_info.account_name
                    "algo_args": {  # 具体的算法参数
                        "order_type": order_type,  # 此柜台不使用下单时要求的算法
                        "start_time": start_time,
                        "end_time": end_time
                    },
                    "status": status,  # 'active' | 'filled' | 'canceled'
                    "filled_vol": filled_vol,  # 实际成交的 volume
                    "dbTime": utc_update_time,
            }
            
            res = replace_collection.insert_one(order_msg)
            self.logger.info(f"[rtn_order] (res){res} (order_msg){order_msg}")


    @MongoClientTradeGateway.error_handler
    def monitor_algo_order_insert(self):
        while not self.is_stopped:
            print ("[monitor_algo_order_insert]")
            for acc in self.accounts_run:
                target_account_name = self.target_account_names[acc]
                sell_query = {"accountName": target_account_name}
                sell_targets = self.order_info_db[acc]["short_target"].find(sell_query)
                if sell_targets.count() == 0:
                    self.logger.warning(
                        f"[monitor_sell_order] no_sell_target (acc){acc}")
                    continue
                for sell_target in sell_targets:
                    if sell_target['_id'] not in self.sell_order_db_ids:
                        self.sell_order_db_ids.append(sell_target['_id'])
                        self.req_sell_order_insert(sell_target)
                    else:
                        self.logger.warning(f"[monitor_algo_order_insert] _id_existed (_id){sell_target['_id']}")
            time.sleep(self.scan_interval)
    
    @MongoClientTradeGateway.error_handler
    def monitor_cancel_order_insert(self):
        while not self.is_stopped:
                for acc in self.accounts_run:
                    print ("[monitor_cancel_order_insert]")
                    target_account_name = self.target_account_names[acc]
                    query = {"accountName": target_account_name}
                    with self.cancel_orderlock:
                        cancel_targets = self.order_info_db[acc]["cancel_target"].find(
                            query)
                        if cancel_targets.count() == 0:
                            self.logger.warning(
                                "[monitor_cancel_order] no_cancel_info")
                            continue
                        for cancel_target in cancel_targets:
                            if cancel_target['_id'] not in self.cancel_order_ids:
                                self.cancel_order_ids.append(cancel_target['_id'])
                                self.req_cancel_order(cancel_target)
                time.sleep(self.scan_interval)

    @MongoClientTradeGateway.error_handler
    def req_cancel_order(self, cancel_target):
        oid = str(cancel_target['oid'])
        mudan_id = ""
        if oid not in self.oid_to_ref:
           self.logger.error("[req_cancel_order] can't_find_mudanid (oid){oid}")
           return
        mudan_id = self.oid_to_ref[oid]
        cxltype = 1
        row = [mudan_id, cxltype]
        self.cancel_order_table.append(row)
        
        origin_req = self.oid_to_req[oid]
        target_account_name = origin_req['accountName']
        acc = self.target_account_names_to_acc[target_account_name]
        target_collection =  self.order_info_db[acc]['cancel_target']
        delete_query = {
            'accountName' : target_account_name,
            'oid' : oid
            }
        delete_res = target_collection.delete_one(delete_query)
        self.logger.info(f"[on_req_cancel_order] delete (target){cancel_target}")
    
    #得到的数据 类似于子单的rtn_order
    @MongoClientTradeGateway.error_handler
    def monitor_trade_update(self):
        fields = ["下单时间", "产品名称", "投资组合名称", "产品Id", "资产单元Id", "投资组合Id", "合同号", "交易类型", "发送标志", "市场",
          "证券代码", "证券名称", "成交状态", "买卖方向", "指令数量", "价格类型", "指令价格", "批次号", "订单类型", "指令金额",
          "策略参数", "合约价格", "成交数量", "成交金额", "撤单数量", "录入用户", "备注", "计价币种", "交易汇率", "用户类型",
          "是否DFD", "交易日期", "是否计算佣金", "合约类型"]

        while not self.is_stopped:
            with self.trade_lock:
                trade_filename = self.recv_msg_dir + '\\' + FILES.TRADES  + '.txt'
                data_list = []

                with open(trade_filename, 'r', encoding=check_charset(trade_filename)) as file:

                    # Read each line in the file
                    for line in file:
                        # Skip lines starting with '#'
                        if line.startswith('#'):
                            continue

                        # Remove leading/trailing whitespaces and split the data using '|'
                        values = line.strip().split('|')


                        row_dict = {fields[i]: values[i] for i in range(len(fields))}

                        # Add the row dictionary to the list
                        data_list.append(row_dict)

                        for record in data_list:

                            if record['发送标志'] in ['已接收'] and not record['成交状态'] in ['未成交']: #only need filled/canceled
                                self.on_trade_msg(record)
                                self.update_order(record)
                            elif record['发送标志'] in ['未接收', '已拒绝']:
                                self.logger.info(f"[monitor_trade_update] trade_not_finished_or_rejected msg{record}")


            time.sleep(30)

    @MongoClientTradeGateway.error_handler
    def on_trade_msg(self, record):
        if record['成交数量'] == 0 :
            self.logger.warning(f"[on_trade_msg] nothing_traded (record){record}")
            return
        oid = str(record['备注'])
        mudan_id = str(record['投资组合Id'])

        acc = ""
        mid = ""
        order_dict = {}
        if not oid in self.oid_to_acc: #shut_down or other problem
            trade_acc = str(record['资产单元Id'])
            acc = self.account_id_to_acc[trade_acc]
            order_info_collection = self.order_info_db[acc]['atx_order']
            query = {'oid' : oid}
            order_info_target = order_info_collection.find_one(query)

            if not order_info_target is None:
                order_dict = order_info_target['order_msg']
                mid = order_info_target['mid']
                self.oid_to_local_ids[oid] = order_info_target['local_ids']
                self.oid_to_traded[oid] = 0
                self.oid_to_traded_money[oid] = 0
                self.oid_to_mid[oid] = mid
                self.oid_to_req[oid] = order_dict
                self.oid_to_ref[oid] = mudan_id
            else:
                self.logger.error(f"[on_trade_msg] can't_find_oid (oid){oid}")
                return
        else:
            acc = self.oid_to_acc[oid]
            order_dict = self.oid_to_req[oid]
            mid = self.oid_to_mid[oid]
        _id = str(record['合同号'])
        if _id in self.oid_to_local_ids[oid]:
            #self.logger.warning(f"[on_trade_msg] duplicate_trade_msg (record){record}")
            return
        else:
            self.oid_to_local_ids[oid].append(_id)
        trade_ref = str(record['合同号']) #外部委托编号，出现在这个位置主要是凑数
        ticker = str(record['证券代码'])
        exchange = decode_exchange_id(ticker)
        traded_vol = int(record['成交数量'])
        traded_price = int(record['成交金额']) / traded_vol
        trade_amt = int(record['成交金额'])

        target_type = encode_ims_side(record['买卖方向'])
        if target_type == 1 or target_type == 4:
            self.oid_to_traded_money[oid] += trade_amt
            self.oid_to_traded[oid] += traded_vol
        elif target_type == 2 or target_type == 3:
            self.oid_to_traded_money[oid] -= trade_amt
            self.oid_to_traded[oid] -= traded_vol
        entrust_vol = record['指令数量']
        #entrust_price = record.Price
        entrust_price = float(str(record['指令价格']).strip(' '))
        dbTime = datetime.now(timezone.utc)

        commission = 0
        #str_trade_time = str(int(int(record['成交时间'])/1000))
        #trade_time = datetime.strptime(str_trade_time, "%Y%m%d%H%M%S")
        #utc_trade_time = trade_time - timedelta(hours=8)
        order_type = order_dict['order_type']
        target_account_name = order_dict['accountName']
        acc = self.target_account_names_to_acc[target_account_name]
        log_account_name = order_dict['accountName'].split('@')[1]
        tg_name =  order_dict['accountName'].split('@')[0]

        side = encode_ims_side(record['买卖方向'])

        db_msg = {
                    "trade_ref": _id, # 用的是非凸的子单order_id 经测试之前用的trade_ref 经常为空
                    "oid": oid,
                    "tg_name": tg_name,
                    "exchange": exchange,
                    "ticker": ticker,
                    "traded_vol": traded_vol,
                    "traded_price": traded_price,
                    "order_type": order_type,
                    "side": side,  # 对应 tlclient.trader.constant 包 Side
                    "entrust_vol": entrust_vol,
                    "entrust_price": entrust_price,
                    "dbTime": dbTime,
                    "mid": mid,  # 对应订单中的 mid
                    "commission": 0,  # 没有
                    "trade_time": dbTime,  # 具体交易时间

                    "accountName":  log_account_name,  # 对应 account_info.account_name
                }
        trade_collection = self.tradelog_db[acc]['trade']
        replace_trade_query = { "trade_ref": _id, "oid" : oid, "mid": mid, "accountName": log_account_name}
        db_res = trade_collection.replace_one(replace_trade_query, db_msg, True)
        self.logger.info(
            f"[rtn_trade] (db_res){db_res} (db_msg){db_msg} (traded_vol){traded_vol} (traded_price){traded_price}")
        self.update_position(traded_vol, order_dict, mid, oid, side, trade_amt,exchange)

        atx_order_collection = self.order_info_db[acc]['atx_order']
        atx_order_msg = {
                    "local_ids" : self.oid_to_local_ids[oid],
                }
        update_atx_order_msg = {"$set": atx_order_msg}
        query = {"oid": oid}
        res = atx_order_collection.update_one(query, update_atx_order_msg)
        self.logger.info(f"[on_trade_msg] update_atx_order_info (res){res} (msg){update_atx_order_msg}")

    @MongoClientTradeGateway.error_handler
    def update_position(self, traded_vol, order, mid , oid, side, trade_amt, exchange):
        target_account_name = order['accountName']
        acc = self.target_account_names_to_acc[target_account_name]

        query = {'mid': mid, 'accountName': target_account_name}
        position_collection = self.order_info_db[acc]['EquityPosition']
        position = position_collection.find_one(query)
        if position is None and traded_vol > 0:
            if oid not in self.oid_to_req:
                self.logger.warning(
                    f"[update_position] can't_find_req (mid){mid} (oid){oid}")
                return
            if oid not in self.oid_to_traded:
                self.logger.error(f"[update_position]no trade_vol (mid){mid}")
                return
            td_trade_vol = self.oid_to_traded[oid]

            amt = self.oid_to_traded_money[oid]

            ticker = order['ticker']
            yd_pos = 0
            mkt_value = amt
            cost = 0
            if not td_trade_vol == 0:
                cost = amt / td_trade_vol
            update_time = datetime.now(timezone.utc)

            position_msg = {
                "ticker": ticker,
                "cost": cost,  # 成本价
                "td_pos_long": td_trade_vol,  # 今仓
                "yd_pos_long": yd_pos,  # 昨仓（可卖），挂止盈单后为0
                "td_pos_short": 0,
                "yd_pos_short": 0,
                "actual_td_pos_long": td_trade_vol,  # 实际的今仓
                "actual_yd_pos_long": yd_pos,  # 实际的昨仓
                "enter_date": update_time,
                "holding_days": 0,  # 持仓时间（交易日累计）
                "mkt_value": mkt_value,  # 市值
                "exchange": exchange,# 对应 tlclient.trader.constant 包 ExchangeID
                "mid": mid,  # 我们策略生成的id，需要根据实际下单情况维护各sid的持仓。
                # 格式: {account_info.tg_name}@{account_info.account_name}
                "accountName": target_account_name,
                "update_date": update_time,
            }
            query2 = {'mid': mid,
                        'accountName': target_account_name}
            res = position_collection.replace_one(query2, position_msg, True)
            self.logger.info(
                f"[update_position] (res){res} (msg){position_msg}")
        else:
            if oid not in self.oid_to_req:
                self.logger.warning(
                    f"[update_position] can't_find_req (mid){mid}")
                return
            if oid not in self.oid_to_traded:
                self.logger.error(f"[update_position]no trade_vol (mid){mid}")
                return
            actual_yd_pos = position['actual_yd_pos_long']
            amt = trade_amt
            yd_pos = position['yd_pos_long']
            if side == 1 or side == 3:
                td_trade_vol = position['actual_td_pos_long'] + traded_vol
                mkt_value = position['mkt_value'] + amt
            elif side == 2 :
                td_trade_vol = position['actual_td_pos_long'] - traded_vol
                mkt_value = position['mkt_value'] - amt
                yd_pos = yd_pos - traded_vol

            ticker = order['ticker']

            cost = 0
            total_vol = td_trade_vol + yd_pos
            if not total_vol == 0:
                cost = mkt_value / total_vol
            update_time = datetime.now(timezone.utc)
            accountName = order['accountName']
            enter_time = position['enter_date']
            holding_days = position['holding_days']
            position_msg = {
                "ticker": ticker,
                "cost": cost,  # 成本价
                "td_pos_long": td_trade_vol,  # 今仓
                "yd_pos_long": yd_pos,  # 昨仓（可卖），挂止盈单后为0
                "td_pos_short": 0,
                "yd_pos_short": 0,
                "actual_td_pos_long": td_trade_vol,  # 实际的今仓
                "actual_yd_pos_long": actual_yd_pos,  # 实际的昨仓
                "enter_date": enter_time,
                "holding_days": holding_days,  # 持仓时间（交易日累计）
                "mkt_value": mkt_value,  # 市值
                "exchange": exchange,  # 对应 tlclient.trader.constant 包 ExchangeID
                "mid": mid,  # 我们策略生成的id，需要根据实际下单情况维护各sid的持仓。
                # 格式: {account_info.tg_name}@{account_info.account_name}
                "accountName": accountName,
                "update_date": update_time,
            }
            query = {'mid': mid,
                        'accountName': accountName}
            res = position_collection.replace_one(query, position_msg, True)
            self.logger.info(
                f"[update_position] (res){res} (msg){position_msg}")

    def start(self):
        signal.signal(signal.SIGTERM, self.signal_handler)
        signal.signal(signal.SIGINT, self.signal_handler)
        _msg = f"[login] ims_server_start (time){datetime.now()}"
        self.send_to_user(logger=self.logger, url_list=self.url_list, msg=_msg)
        self.monitor()

    def close(self):
        return super().close()
    
    def join(self):
        while self.is_stopped == False:
            time.sleep(0.01)
            if self.is_stopped:
                self.logger.info(
                    "[close] main thread is stopped,active_orders message will lose")
                self.close()

if __name__ == "__main__":

    sys.stdout.reconfigure(encoding='utf-8')
    description = "atx_server,get target from mongodb and serve atx"  
    parser = argparse.ArgumentParser(description=description)

    parser.add_argument('-e' , '--end_time', dest='end_time', default='15:30')
    _config_filename = "C:/Users/Administrator/Desktop/ims_batandconfig/ims_server_config.json"
    parser.add_argument('-p', '--config_filepath', dest= 'config_filepath', default= _config_filename)

    args = parser.parse_args()
    print (f"(args){args}")

    td = ImsServer(args.config_filepath, args.end_time)
    td.start()

    td.join()




class ORDER_SIDE:
    BUY  = 1
    SELL = 2
    MB =  11 #融资买入
    RS =  22 #卖券还款
    SS =  12 #融券卖出
    RB =  21 #买券还券

    
class DATA_STATUS:
    UNSUBMITTED : 0 #初始化未完成
    SUBMITTED : 1   #报单插入柜台成功
    STOP : 2 #暂停
    FILLED : 3 #完成
    CANCELLED : 4 #取消
    FAILED : 6 #错单废单


ENCODING = 'gb2312'
GATEWAY_NAME = 'cats_uni'


class FILES:
    TASKS = 'instructions'  #下单
    CANCEL = 'instructions'
    ORDERS = 'order_updates'
    ALGO_ORDERS = 'stats_all'
    TRADES = 'trade'
    POSITIONS = 'asset'


# encoding: utf-8

from mquant_api import *
from mquant_struct import *
import json
import os
import time
import threading
import shutil
import datetime


def transfer_order_side(side):
    if side == OrderSide.BUY:
        return 1
    elif side == OrderSide.SELL:
        return 2
    else:
        return -1


def transfer_price_type(style):
    if isinstance(style, LimitOrderStyle) or isinstance(style, MarketOrderStyle):
        return style.get_order_style()
    else:
        return 'a'


def get_entrust_price(style):
    if isinstance(style, LimitOrderStyle) or isinstance(style, MarketOrderStyle):
        return style.get_limited_price()
    else:
        return 0


class TradeDataDownloader(object):
    """
    数据下载类
    """

    def __init__(self):
        self.__context = None
        cur_date = datetime.datetime.strftime(datetime.datetime.now(), "%Y-%m-%d")
        self.__file_dir = 'D\\MQuantTradeData\\' + cur_date
        self.__custom_field_mutex = threading.Lock()
        self.__complete_section_mutex = threading.Lock()
        self.__order_custom_field = {}
        self.__complete_section = False
        self.__enable_custom_field = False
        self.__order_req_info = {}
        self.__account_type_info = {}
        self.__bak_file_num = 0
        self.__only_download_mquant_order = False
        self.write_pos_lock = threading.Lock()

    def set_complete_section(self, flag):
        self.__complete_section = flag

    def set_enable_custom_field(self, flag):
        self.__enable_custom_field = flag

    def set_bak_file_num(self, num):
        self.__bak_file_num = num

    def set_context(self, context):
        self.__context = context
        fund_account_stock = context.get_fund_account_by_type(AccountType.normal)
        if fund_account_stock is not None and len(fund_account_stock) > 0:
            self.__account_type_info[fund_account_stock] = AccountType.normal

        fund_account_margin = context.get_fund_account_by_type(AccountType.margin)
        if fund_account_margin is not None and len(fund_account_margin) > 0:
            self.__account_type_info[fund_account_margin] = AccountType.margin

    def set_file_dir_path(self, dir_path):
        if not os.path.exists(dir_path):
            os.mkdir(dir_path)
        cur_date = datetime.datetime.strftime(datetime.datetime.now(), "%Y-%m-%d")
        self.__file_dir = dir_path + '\\' + cur_date
        if not os.path.exists(self.__file_dir):
            os.mkdir(self.__file_dir)

    def only_download_mquant_order(self, flag):
        self.__only_download_mquant_order = flag


    def get_order_file_path(self, fund_account):
        """
        获取订单文件路径
        :param fund_account:
        :return:
        """
        if fund_account is None:
            return ''
        if len(fund_account) == 0:
            return ''
        return self.__file_dir + '\\order_' + fund_account + '.csv'

    def get_execution_file_path(self, fund_account):
        """
        获取成交文件路径
        :param fund_account:
        :return:
        """
        if fund_account is None:
            return ''
        if len(fund_account) == 0:
            return ''
        return self.__file_dir + '\\execution_' + fund_account + '.csv'

    def get_fund_file_path(self, fund_account):
        """
        获取资金文件路径
        :param account_type:
        :return:
        """
        if fund_account is None:
            return ''
        if len(fund_account) == 0:
            return ''
        return self.__file_dir + '\\fund_' + fund_account + '.csv'

    def get_pos_file_path(self, fund_account):
        """
        获取持仓文件路径
        :param fund_account:
        :return:
        """
        if fund_account is None:
            return ''
        if len(fund_account) == 0:
            return ''
        return self.__file_dir + '\\position_' + fund_account + '.csv'

    def get_custom_field_file_path(self):
        """
        获取自定义字段文件路径
        :return:
        """
        return self.__file_dir + '\\custom_field.csv'

    def init_trade_files(self):
        """
        交易数据初始化
        :return:
        """
        self.init_custom_fields()
        self.init_order_files(AccountType.normal)
        self.init_order_files(AccountType.margin)
        self.init_execution_files(AccountType.normal)
        self.init_execution_files(AccountType.margin)
        self.init_fund_files(AccountType.normal)
        self.init_pos_files(AccountType.normal)
        self.init_fund_files(AccountType.margin)
        self.init_pos_files(AccountType.margin)

    def init_custom_fields(self):
        """
        初始化自定义字段
        :return:
        """
        if not self.__enable_custom_field:
            return
        custom_field_file_path = self.get_custom_field_file_path()
        if not os.path.exists(custom_field_file_path):
            with open(custom_field_file_path, 'w', encoding='utf-8') as f:
                f.write('订单号,自定义字段\n')
        else:
            # 将自定义字段信息都读取出来
            with open(custom_field_file_path, 'r', encoding='utf-8') as f:
                info_list = f.readlines()
                if len(info_list) > 0:
                    while not self.__custom_field_mutex.acquire():
                        time.sleep(0.01)
                    for line in info_list:
                        lst_item = line.rstrip('\n').split(',')
                        self.__order_custom_field[lst_item[0]] = lst_item[1]
                    self.__custom_field_mutex.release()

    def init_order_files(self, account_type):
        """
        初始化订单数据
        :param account_type:
        :return:
        order_info = '{},{},{},{},{},{},{},{},{},{},{},{},{},{},{},{},{},{},{},{},{},{},{}\n'.format(ord.fund_account, ord.add_time,
                                                                                         ord.inst_id, ord.symbol,
                                                                                         transfer_order_side(ord.side),
                                                                                         ord.amount,
                                                                                         get_entrust_price(ord.style),
                                                                                         ord.status.value,
                                                                                         transfer_price_type(ord.style),
                                                                                         ord.entrust_no,
                                                                                         ord.entrust_prop,
                                                                                         ord.entrust_type,
                                                                                         ord.filled,
                                                                                         ord.business_balance,
                                                                                         ord.price,
                                                                                         ord.withdraw_amount,
                                                                                         cancel_info, ord.order_id,
                                                                                         custom_field, algo_inst_id,
                                                                                        datetime.datetime.now(),
                                                                                        ord.security_exchange,
                                                                                        ord.batch_no)
    }
        """
        order_file = self.get_order_file_path(self.__context.get_fund_account_by_type(account_type))
        if len(order_file) == 0:
            return

        with open(order_file, 'w', encoding='utf-8') as f:
            f.write(
                'fund_account,add_time,inst_id,symbol,side,entrust_vol,entrust_price,status,price_type,entrust_no,entrust_prop,entrust_type,filled,business_balance,price,withdraw_amount,cancel_info,order_id,custom_field,algo_inst_id,update_time,market,batch_no\n')
            page_no = 1
            tmp_orders = []
            while True:
                ret_data = get_orders_ex(only_this_inst=False, page_no=page_no,
                                         account_type=account_type)  # 仅演示获取普通A股账号的订单信息，可根据自己的实际场景选择获取两融的,分页查询，每次获取1000条
                ord_list = ret_data[2]
                if ord_list is not None:
                    tmp_orders = tmp_orders + list(ord_list.values())
                    
                    if ret_data[1]:
                        break
                    page_no = page_no + 1
                else:
                    break
            tmp_orders = sorted(tmp_orders, key=lambda ord: str(ord.add_time) + str(ord.entrust_no))
#            pre_time = None
            for ord in tmp_orders:
                if self.__only_download_mquant_order and (ord.inst_id is None or len(ord.inst_id) == 0):
                    continue
#                if pre_time is not None:
#                    if ord.add_time < pre_time:
#                        print('sort error', ord.add_time, pre_time)
#                pre_time = ord.add_time
#                log.debug('{}'.format(ord.add_time))
                f.write(self.get_order_str(ord))

    def init_execution_files(self, account_type):
        """
        初始化成交文件
        :param account_type:
        :return:
        """
        execution_file = self.get_execution_file_path(self.__context.get_fund_account_by_type(account_type))
        if len(execution_file) == 0:
            return
        with open(execution_file, 'w', encoding='utf-8') as f:
            f.write('fund_account,time,inst_id,symbol,side,price,amount,business_balance,real_type,entrust_no,trade_id,order_id,custom_field,algo_inst_id,update_time,market\n')
            page_no = 1
            tmp_execution_list = []
            while True:
                ret_data = get_trades_ex(only_this_inst=False, page_no=page_no, page_size=1000,
                                         account_type=account_type)  # 仅演示获取普通A股账号的订单信息，可根据自己的实际场景选择获取两融的

                execution_list = ret_data[2]
                if execution_list is not None:
                    print('查询成交返回：', len(execution_list))
                    tmp_execution_list = tmp_execution_list + execution_list
                    

                    if ret_data[1]:
                        break
                    page_no = page_no + 1
                else:
                    break
            tmp_execution_list = sorted(tmp_execution_list,
                                            key=lambda execution: str(execution.time) + str(execution.trade_id))

            for execution in tmp_execution_list:
                if self.__only_download_mquant_order and (execution.inst_id is None or len(execution.inst_id) == 0):
                    continue
                f.write(self.get_execution_str(execution))

    def init_fund_files(self, account_type):
        """
        初始化资金文件
        :return:
        """
        fund_account = self.__context.get_fund_account_by_type(account_type)
        fund_file = self.get_fund_file_path(fund_account)
        if len(fund_file) == 0:
            return
        with open(fund_file, 'w', encoding='utf-8') as f:
            f.write(self.get_fund_str(fund_account))

    def init_pos_files(self, account_type):
        """
        初始化持仓信息
        :return:
        """
        # 持仓
        self.write_pos_info(self.__context.get_fund_account_by_type(account_type))

    def get_bak_file(self, origin_file, subfix=0):
        """

        :param origin_file:
        :param subfix:
        :return:
        """
        #    print(origin_file, subfix, g.bak_file_num)
        if subfix > g.bak_file_num:
            return ""
        pos = origin_file.find('.csv')
        if pos > 0:
            prefix = origin_file[:pos]
            file_bak = prefix + '_bak' + str(subfix) + '.csv'
            print(file_bak)
            if os.path.exists(file_bak):
                return get_bak_file(origin_file, subfix + 1)
            else:
                return file_bak
        return ""

    def bak_trade_files(self, order_file, execution_file, fund_file, pos_file):
        """

        :param order_file:
        :param execution_file:
        :param fund_file:
        :param pos_file:
        :return:
        """

        if os.path.exists(order_file):
            ord_bak_file = get_bak_file(order_file)
            if len(ord_bak_file) > 0:
                shutil.copy(order_file, ord_bak_file)

        if os.path.exists(execution_file):
            exec_bak_file = get_bak_file(execution_file)
            if len(exec_bak_file) > 0:
                shutil.copy(execution_file, exec_bak_file)


    def write_fund_info(self, context, account_type, fund_file):
        """
        写资金
        :param context:
        :param account_type:
        :param fund_file:
        :return:
        """
        with open(fund_file, 'a', encoding='utf-8') as f:
            context.set_current_account_type(account_type)
            fund_account = context.get_fund_account_by_type(account_type)
            fund_str = '{},{},{},{}\n'.format(datetime.datetime.now(), fund_account, context.portfolio.settled_cash,
                                              context.portfolio.available_cash)
            f.write(fund_str)

    def write_pos_info(self, fund_account):
        """
        写持仓
        :param account_type:
        :param pos_file:
        :return:
        """
        if fund_account is None or len(fund_account) == 0:
            return
        pos_file = self.get_pos_file_path(fund_account)
        pos_str = 'fund_account,symbol,total_amount,closeable_amount,init_amount,holding_cost,value,update_time,market\n'

        account_type = self.__account_type_info[fund_account]
        pos_list = get_positions_ex(account_type)
        if pos_list is not None:
            for pos in pos_list:
                pos_str = pos_str + '{},{},{},{},{},{},{},{},{}\n'.format(fund_account, pos.security, pos.total_amount,
                                                        pos.closeable_amount,
                                                        pos.init_amount,pos.hold_cost, pos.value,datetime.datetime.now(),pos.security_exchange)
        else:
            print('get_positions_ex return None')
            return
        with self.write_pos_lock:
            with open(pos_file, 'w', encoding='utf-8', buffering=len(pos_str) + 1) as f:
                f.write(pos_str)

    def record_order(self, order_id, symbol, side, amount, price, price_type):
        """
        记录订单请求
        :param order_id:
        :param symbol:
        :param side:
        :param amount:
        :param price:
        :param price_type:
        :return:
        """
        #    print('record', price, price_type)
        if not self.__complete_section:
            return
        while not self.__complete_section_mutex.acquire():
            time.sleep(0.001)
        if len(order_id) > 0:
            self.__order_req_info[order_id] = {'symbol': symbol, 'side': side, 'amount': amount, 'price': price,
                                               'price_type': price_type,
                                               'time': datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')}
        self.__complete_section_mutex.release()

    def get_order_ext_info(self, order_id):
        """
        获取订单请求信息
        :param order_id:
        :return:
        """
        if not self.__complete_section:
            return
        while not self.__complete_section_mutex.acquire():
            time.sleep(0.001)
        if len(order_id) > 0 and self.__order_req_info.get(order_id) is not None:
            ext_order_info = self.__order_req_info[order_id]
            self.__complete_section_mutex.release()
            return ext_order_info
        self.__complete_section_mutex.release()

    def get_custom_field_str(self, ord, push_msg=False):
        """
        获取自定义字段
        """
        if not self.__enable_custom_field:
            return ""
        # 排除非MQuant报单
        if ord.inst_id is None or len(ord.inst_id) == 0:
            #        print('not mquant order:', ord.order_id)
            return ""
        # 排除算法报单
        if ord.algo_inst_id is not None and len(ord.algo_inst_id) > 0:
            return ""

        while not self.__custom_field_mutex.acquire():
            time.sleep(0.01)

        custom_field = self.__order_custom_field.get(ord.order_id)
        self.__custom_field_mutex.release()
        count = 0
        if push_msg:
            while custom_field is None:
                time.sleep(0.01)
                while not self.__custom_field_mutex.acquire():
                    time.sleep(0.001)
                custom_field = self.__order_custom_field.get(ord.order_id)
                self.__custom_field_mutex.release()
                count = count + 1
                if count > 10:
                    custom_field = ''
                    log.warn('尝试获取100次订单的用户自定义字段失败，订单id：{}'.format(ord.order_id))
                    break
        elif custom_field is None:
            custom_field = ''
        return custom_field

    def get_custom_field_str_from_execution(self, execution):
        if not self.__enable_custom_field:
            return ""
        if execution.inst_id is None or len(execution.inst_id) == 0:
            #        print('not mquant order:', ord.order_id)
            return ""
        if execution.algo_inst_id is not None and len(execution.algo_inst_id) > 0:
            return ""

        while not self.__custom_field_mutex.acquire():
            time.sleep(0.001)

        custom_field = self.__order_custom_field.get(execution.order_id)
        self.__custom_field_mutex.release()
        if custom_field is None:
            custom_field = ''
        return custom_field

    def record_custom_field(self, order_id, custom_field):
        """
        记录自定义字段
        """
        if not self.__enable_custom_field:
            return
        while not self.__custom_field_mutex.acquire():
            time.sleep(0.001)
        self.__order_custom_field[order_id] = custom_field
        self.__custom_field_mutex.release()

        # 将自定义字段写入文件
        try:
            with open(self.get_custom_field_file_path(), 'a', encoding='utf-8') as f:
                #            print('record custom field:', order_id, custom_field)
                f.write('{},{}\n'.format(order_id, custom_field))
        except Exception as e:
            print(e)

    def get_order_str(self, ord, push_msg=False):
        """
        订单字符串
        :param ord:
        :param push_msg:
        :return:
        """
        custom_field = ''
        #    if push_msg:
        custom_field = self.get_custom_field_str(ord, push_msg)
        #    print(ord.order_id, ord.status, ord.fund_account)
        if ord.status is None:
            log.debug(ord.order_id)
        cancel_info = ord.cancel_info
        if len(ord.cancel_info) > 0:
            cancel_info = ord.cancel_info.replace(",", ";").replace("\r\n", "")
            cancel_info = cancel_info.replace(",", ";").replace("\n", "")
        if (len(ord.symbol) == 0 or ord.status == OrderStatus.rejected) and self.__complete_section:
            ord_ext_info = self.get_order_ext_info(ord.order_id)
            if ord_ext_info is not None:
                ord.symbol = ord_ext_info['symbol']
                ord.amount = ord_ext_info['amount']
                price = ord_ext_info['price']
                ord.side = ord_ext_info['side']
                ord.add_time = ord_ext_info['time']
                price_type = ord_ext_info['price_type']
                if str(price_type) == "1":
                    ord.style = LimitOrderStyle(price)
                else:
                    ord.style = MarketOrderStyle(price_type, price)

        algo_inst_id = ord.algo_inst_id
        if algo_inst_id is None:
            algo_inst_id = ''
        order_info = '{},{},{},{},{},{},{},{},{},{},{},{},{},{},{},{},{},{},{},{},{},{},{}\n'.format(ord.fund_account, ord.add_time,
                                                                                         ord.inst_id, ord.symbol,
                                                                                         transfer_order_side(ord.side),
                                                                                         ord.amount,
                                                                                         get_entrust_price(ord.style),
                                                                                         ord.status.value,
                                                                                         transfer_price_type(ord.style),
                                                                                         ord.entrust_no,
                                                                                         ord.entrust_prop,
                                                                                         ord.entrust_type,
                                                                                         ord.filled,
                                                                                         ord.business_balance,
                                                                                         ord.price,
                                                                                         ord.withdraw_amount,
                                                                                         cancel_info, ord.order_id,
                                                                                         custom_field, algo_inst_id,
                                                                                        datetime.datetime.now(),
                                                                                        ord.security_exchange,
                                                                                        ord.batch_no)
        return order_info

    def get_execution_str(self, execution, push_msg=False):
        """
        成交字符串
        :param push_msg:
        :return:
        
        """
        custom_field = self.get_custom_field_str_from_execution(execution)
        algo_inst_id = execution.algo_inst_id
        if algo_inst_id is None:
            algo_inst_id = ''
        # '资金账号,成交时间,实例ID,证券代码,买卖方向,成交价格,成交数量,成交金额,成交类型,柜台委托编号,柜台成交编号,订单编号,自定义字段'
        execution_info = '{},{},{},{},{},{},{},{},{},{},{},{},{},{},{},{}\n'.format(execution.fund_account, execution.time,
                                                                              execution.inst_id, execution.symbol,
                                                                              transfer_order_side(execution.side),
                                                                              execution.price,
                                                                              execution.amount,
                                                                              execution.business_balance,
                                                                              execution.real_type,
                                                                              execution.entrust_no, execution.trade_id,
                                                                              execution.order_id,
                                                                              custom_field, algo_inst_id,
                                                                            datetime.datetime.now(),
                                                                            execution.security_exchange)

        return execution_info

    def get_fund_str(self, fund_account, fund_info=None):
        """
        获取资金串
        :param fund_account:
        :param fund_info:
        :return:
        """
        # if fund_info is None:
        #     fund_str = '记录时间,资金账号,期初资金,可用资金\n' + '{},{},{},{}\n'.format(
        #         datetime.datetime.now(),
        #         fund_account,
        #         self.__context.portfolio.settled_cash,
        #         self.__context.portfolio.available_cash)
        # else:
        #     fund_str = '记录时间,资金账号,期初资金,可用资金\n' + '{},{},{},{}\n'.format(
        #         datetime.datetime.now(),
        #         fund_account,
        #         fund_info.settled_cash,
        #         fund_info.available_cash)
        # return fund_str

        if fund_info is None:
            fund_str = 'write_time,fund_account,settled_cash,available_cash,total_cash,frozen_cash,transferable_cash,market_value,total_value,hk_available_cash\n' + '{},{},{},{},{},{},{},{},{},{}\n'.format(
                datetime.datetime.now(),
                fund_account,
                self.__context.portfolio.settled_cash,
                self.__context.portfolio.available_cash,
                self.__context.portfolio.total_cash,
                self.__context.portfolio.frozen_cash,
                self.__context.portfolio.transferable_cash,
                self.__context.portfolio.market_value,
                self.__context.portfolio.total_value,
                self.__context.portfolio.hk_available_cash)
        else:
            fund_str = 'write_time,fund_account,settled_cash,available_cash,total_cash,frozen_cash,transferable_cash,market_value,total_value,hk_available_cash\n' + '{},{},{},{},{},{},{},{},{},{}\n'.format(
                datetime.datetime.now(),
                fund_account,
                fund_info.settled_cash,
                fund_info.available_cash,
                fund_info.hold_cash,
                fund_info.frozen_cash,
                fund_info.transferable_cash,
                fund_info.market_value,
                fund_info.total_value,
                fund_info.hk_available_cash)
        return fund_str

    def write_order_to_csv(self, ord):
        """
        写订单
        :return:
        """
        file_path = self.get_order_file_path(ord.fund_account)
        if not os.path.exists(file_path):
            log.error("订单文件{}不存在，无法更新订单信息".format(file_path))
            return
        with open(file_path, 'a', encoding='utf-8') as f:
            # 字段：资金账号、委托时间、实例ID、证券代码、买卖方向、委托数量、委托价格、价格类型、柜台委托编号、委托属性、委托类型、成交数量、成交金额、撤单数量、状态、废单原因、自定义字段
            f.write(self.get_order_str(ord, True))
            
    def write_order_list_to_csv(self, ord_list):
        """
        写订单
        :return:
        """
        file_path = self.get_order_file_path(ord.fund_account)
        if not os.path.exists(file_path):
            log.error("订单文件{}不存在，无法更新订单信息".format(file_path))
            return
        with open(file_path, 'a', encoding='utf-8') as f:
            # 字段：资金账号、委托时间、实例ID、证券代码、买卖方向、委托数量、委托价格、价格类型、柜台委托编号、委托属性、委托类型、成交数量、成交金额、撤单数量、状态、废单原因、自定义字段
            for ord in ord_list:
                f.write(self.get_order_str(ord, True))       

    def write_execution_to_csv(self, execution):
        """
        写成交
        :param execution:
        :return:
        """
        # 成交留待客户自行实现
        file_path = self.get_execution_file_path(execution.fund_account)
        if not os.path.exists(file_path):
            log.error("成交文件{}不存在，无法更新订单信息".format(file_path))
            return
        with open(file_path, 'a', encoding='utf-8') as f:
            # 字段：资金账号、委托时间、实例ID、证券代码、买卖方向、委托数量、委托价格、价格类型、柜台委托编号、委托属性、委托类型、成交数量、成交金额、撤单数量、状态、废单原因
            f.write(self.get_execution_str(execution, True))
    
    def write_execution_list_to_csv(self, execution_list):
        """
        写成交
        :param execution:
        :return:
        """
        # 成交留待客户自行实现
        file_path = self.get_execution_file_path(execution.fund_account)
        if not os.path.exists(file_path):
            log.error("成交文件{}不存在，无法更新订单信息".format(file_path))
            return
        with open(file_path, 'a', encoding='utf-8') as f:
            # 字段：资金账号、委托时间、实例ID、证券代码、买卖方向、委托数量、委托价格、价格类型、柜台委托编号、委托属性、委托类型、成交数量、成交金额、撤单数量、状态、废单原因
            for execution in execution_list:
                f.write(self.get_execution_str(execution, True))

    def on_order_update(self, ord):
        """
        订单推送
        :param ord:
        :return:
        """
        if self.__only_download_mquant_order and ( ord.inst_id is None or len(ord.inst_id) == 0):
            return 
        file_path = self.get_order_file_path(ord.fund_account)
        if not os.path.exists(file_path):
            log.error("订单文件{}不存在，无法更新订单信息".format(file_path))
            return
        with open(file_path, 'a', encoding='utf-8') as f:
            # 字段：资金账号、委托时间、实例ID、证券代码、买卖方向、委托数量、委托价格、价格类型、柜台委托编号、委托属性、委托类型、成交数量、成交金额、撤单数量、状态、废单原因、自定义字段
            f.write(self.get_order_str(ord, True))

    def on_execution_update(self, execution):
        """
        成交推送
        :param context:
        :param execution:
        :return:
        """
        if self.__only_download_mquant_order and ( execution.inst_id is None or len(execution.inst_id) == 0):
            return 
        self.write_execution_to_csv(execution)

    def on_fund_update(self, fund_info):
        """
        category:事件回调
        brief:资金推送
        desc:资金推送函数，可选实现。
        :param fund_info: FundUpdateInfo 类对象
        :return:
        """
        fund_file = self.get_fund_file_path(fund_info.fund_account)
        if not os.path.exists(fund_file):
            log.error("资金文件{}不存在，无法更新资金信息".format(fund_file))
            return
        fund_str = self.get_fund_str(fund_info.fund_account, fund_info)
        with open(fund_file, 'w', encoding='utf-8', buffering=len(fund_str) + 1) as f:
            f.write(fund_str)
        self.write_pos_info(fund_info.fund_account)

    def on_position_update(self, pos_info):
        """
        category:事件回调
        brief:持仓推送
        desc:持仓推送函数，可选实现。
        :param pos_info: Position类对象
        :return:
        """
        self.write_pos_info(pos_info.fund_account)

